/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.rackspace.compute;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageFormat;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.MachineImageSupport;
import org.dasein.cloud.compute.MachineImageType;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.jclouds.rackspace.RackspaceCloud;
import org.dasein.util.CalendarWrapper;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;
import org.jclouds.cloudservers.CloudServersAsyncClient;
import org.jclouds.cloudservers.CloudServersClient;
import org.jclouds.cloudservers.domain.Image;
import org.jclouds.cloudservers.domain.ImageStatus;
import org.jclouds.cloudservers.options.ListOptions;
import org.jclouds.rest.RestContext;

public class RackspaceImage implements MachineImageSupport {

	static private final Logger logger = Logger.getLogger(RackspaceImage.class);
    
    private RackspaceCloud provider = null;
    
    public RackspaceImage(RackspaceCloud provider) {
        this.provider = provider;
    }

    @Override
    public void downloadImage(String machineImageId, OutputStream toOutput) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No downloading images");
    }
    
    private CloudServersAsyncClient getAsyncClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getAsyncApi();
    }
    
    private CloudServersClient getClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getApi();
    }
    
    @Override
    public MachineImage getMachineImage(String imageId) throws InternalException, CloudException {
        for( MachineImage img : listMachineImages() ) {
            if( img.getProviderMachineImageId().equals(imageId) ) {
                return img;
            }
        }
        return null;
    }
    
    @Override
    public String getProviderTermForImage(Locale locale) {
        return "image";
    }
    
    @Override
    public boolean hasPublicLibrary() {
        return false;
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachine(String vmId, String name, String description) throws CloudException, InternalException {
        logger.debug("enter - imageServer(String, String, String)");
        try {
            final AsynchronousTask<String> task = new AsynchronousTask<String>();
            final Future<Image> result;

            task.setStartTime(System.currentTimeMillis());

            RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

            try {
                result = getAsyncClient(ctx).createImageFromServer(name, Integer.parseInt(vmId));
            }
            finally {
                ctx.close();
            }
            Thread t = new Thread() {
                public void run() {
                    while( !result.isDone() ) {
                        try { Thread.sleep(5000L); }
                        catch( InterruptedException e ) { }
                    }
                    task.setPercentComplete(90.0);
                    if( result.isCancelled() ) {
                        task.complete(new InterruptedException("Bundle Canceled"));
                    }
                    else {
                        try {
                            String id = String.valueOf(result.get().getId());
                            long timeout = System.currentTimeMillis() + (CalendarWrapper.MINUTE*30L);
                            
                            try { Thread.sleep(1000L); }
                            catch( InterruptedException e ) { }
                            MachineImage image = getMachineImage(id);
                            
                            while( image == null ) {
                                if( System.currentTimeMillis() >= timeout ) {
                                    throw new CloudException("Timeout waiting for image " + id + " to be ready");
                                }
                                try { Thread.sleep(15000L); }
                                catch( InterruptedException e ) { }
                                image = getMachineImage(id);
                            }
                            task.completeWithResult(id);
                        }
                        catch( Exception e ) {
                            logger.error("Bundle of server failed: " + e.getMessage());
                            e.printStackTrace();
                            task.complete(new CloudException(e));
                        }
                    }
                }
            };
            
            t.setDaemon(true);
            t.setName("Create Image - " + name + "/" + vmId);
            t.start();
            return task;
        }
        finally {
            logger.debug("exit - imageServer(String, String, String)");
        }
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public String installImageFromUpload(MachineImageFormat format, InputStream imageStream) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No upload of images supported.");
    }

    @Override
    public boolean isImageSharedWithPublic(String machineImageId) throws CloudException, InternalException {
        return false;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        return true;
    }

    @Override
    public Iterable<MachineImage> listMachineImages() throws CloudException, InternalException {
        PopulatorThread<MachineImage> populator;
        
        provider.hold();
        populator = new PopulatorThread<MachineImage>(new JiteratorPopulator<MachineImage>() {
            public void populate(Jiterator<MachineImage> iterator) throws CloudException, InternalException {
                try {
                    populateImages(iterator);
                }
                finally {
                    provider.release();
                }
            }
        });
        populator.populate();
        return populator.getResult();
    }

    @Override
    public Iterable<MachineImage> listMachineImagesOwnedBy(String accountId) throws CloudException, InternalException {
        if( accountId == null ) {
            return Collections.emptyList();
        }
        else if( accountId.equals(provider.getContext().getAccountNumber()) ) {
            return listMachineImages();
        }
        else {
            return Collections.emptyList();
        }
    }

    @Override
    public Iterable<String> listShares(String forMachineImageId) throws CloudException, InternalException {
        return Collections.emptyList();
    }

    static private List<MachineImageFormat> supportedFormats = null;
    
    @Override
    public Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
        if( supportedFormats == null ) {
            //supportedFormats = Collections.unmodifiableList(Collections.singletonList(MachineImageFormat.OVF));
            supportedFormats = Collections.emptyList();
        }
        return supportedFormats;
    }
    
    private void populateImages(Jiterator<MachineImage> iterator) throws CloudException, InternalException {
        logger.debug("enter - list(String)");
        try {
            Collection<Image> rawImages;
            
            try {
                RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

                try {
                    rawImages = getClient(ctx).listImages(ListOptions.Builder.withDetails());
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Could not list images from Rackspace: " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( Image image: rawImages ) {
                if( image != null ) {
                    iterator.push(toImage(image));
                }
            }
        }
        finally {
            logger.debug("exit - list(String)");
        }
    }
    
    @Override
    public String registerMachineImage(String atStorageLocation) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Rackspace does not support image uploading");
    }
    
	@Override
	public void remove(String imageId) throws InternalException, CloudException {
	    logger.debug("enter - delete(String)");
	    try {
	        RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();
	        
	        ctx.getApi().deleteImage(Integer.parseInt(imageId));
	    }
	    finally {
	        logger.debug("exit - delete(String)");
	    }
	}

	@Override
	public Iterable<MachineImage> searchMachineImages(String keyword, Platform platform, Architecture architecture) throws InternalException,	CloudException {
	    logger.debug("enter - find(String, String, Platform, String)");
	    try {
    		Iterable<MachineImage> images = listMachineImages();
    		Collection<MachineImage> found = new ArrayList<MachineImage>();
    		
    		for (MachineImage image : images) {
    			if (image.getArchitecture().name().equals(architecture) && image.getName().indexOf(keyword) > -1 && image.getPlatform().equals(platform)) {
    				found.add(image);
    			}
    		}
    		
    		return found;
	    }
	    finally {
	        logger.debug("exit - find(String, String, Platform, String)");
	    }
	}

    @Override
    public void shareMachineImage(String machineImageId, String withAccountId, boolean allow) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Rackspace does not support image sharing.");
    }

    @Override
    public boolean supportsCustomImages() {
        return true;
    }

    @Override
    public boolean supportsImageSharing() {
        return false;
    }

    @Override
    public boolean supportsImageSharingWithPublic() {
        return false;
    }
    
    private MachineImage toImage(Image rackspaceImage) throws InternalException, CloudException {
        MachineImage image = new MachineImage();

        image.setProviderRegionId(provider.getContext().getRegionId());
        image.setProviderOwnerId(provider.getContext().getAccountNumber());
        image.setType(MachineImageType.STORAGE);
        image.setName(rackspaceImage.getName());
        image.setDescription(image.getName());
        image.setArchitecture(Architecture.I64);
        image.setPlatform(Platform.guess(image.getName()));
        image.setSoftware("");
        if( image.getPlatform().equals(Platform.UNKNOWN) ) {
            image.setPlatform(Platform.UNIX);
        }
        image.setProviderMachineImageId(String.valueOf(rackspaceImage.getId()));
        
        image.setCurrentState(toState(rackspaceImage.getStatus()));
        return image;
    }
    
    private MachineImageState toState(ImageStatus imageStatus) {
        switch( imageStatus ) {
            case ACTIVE: return MachineImageState.ACTIVE;
            case PREPARING: case QUEUED: case SAVING: case UNRECOGNIZED: return MachineImageState.PENDING;
            case FAILED: return MachineImageState.DELETED;
            default: return MachineImageState.PENDING;
        }   
    }

    @Override
    public String transfer(CloudProvider fromCloud, String machineImageId) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Rackspace does not support image transfer.");
    }
}
