/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.rackspace.platform;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.jclouds.rackspace.RackspaceCloud;
import org.dasein.cloud.jclouds.rackspace.storage.RackspaceCloudFiles;
import org.dasein.cloud.platform.CDNSupport;
import org.dasein.cloud.platform.Distribution;
import org.jclouds.blobstore.BlobStoreContext;
import org.jclouds.cloudfiles.CloudFilesClient;
import org.jclouds.cloudfiles.domain.ContainerCDNMetadata;
import org.jclouds.cloudfiles.options.ListCdnContainerOptions;

public class RackspaceCDN implements CDNSupport {

	static private final Logger logger = Logger.getLogger(RackspaceCDN.class);
    
    private RackspaceCloud provider = null;
    
    RackspaceCDN(RackspaceCloud provider) { this.provider = provider;}
	
	public String create(String directory, String name, boolean active, String ... cnames) throws InternalException, CloudException {
        logger.debug("enter - create(String, String, boolean, String...)");
        try {
            RackspaceCloudFiles blobs = provider.getStorageServices().getBlobStoreSupport();

            try {
                if( !blobs.exists(directory) ) {
                    blobs.createDirectory(directory, false);
                }
            }
            catch( RuntimeException e ) {
                logger.error("Could not create container '" + directory + "': " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            if( active ) {
                BlobStoreContext ctx = provider.getCloudFilesContext();
                
                try {
                    CloudFilesClient client = CloudFilesClient.class.cast(ctx.getProviderSpecificContext().getApi());
                    client.enableCDN(directory);
                }
                catch( RuntimeException e ) {
                    logger.error("Could not enable distribution '" + directory + "': " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
                finally {
                    ctx.close();                        
                }
            }
            return directory;
        }
        finally {
            logger.debug("exit - create(String, String, boolean, String...)");            
        }
	}

	public void delete(String distributionId) throws InternalException, CloudException {
	    logger.debug("enter - delete(String)");
	    try {
            BlobStoreContext ctx = provider.getCloudFilesContext();
            
            try {
                CloudFilesClient client = CloudFilesClient.class.cast(ctx.getProviderSpecificContext().getApi());
                client.disableCDN(distributionId);
            }
            catch( RuntimeException e ) {
                logger.error("Could not disavle distribution '" + distributionId + "': " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            finally {
                ctx.close();                        
            }
	    }
	    finally {
	        logger.debug("exit - delete(String)");
	    }
	}
	    
	public Distribution getDistribution(String distributionId) throws InternalException, CloudException {
	    for( Distribution distribution : list() ) {
	        if( distribution.getProviderDistributionId().equals(distributionId) ) {
	            return distribution;
	        }
	    }
	    return null;
	}

	public String getProviderTermForDistribution(Locale locale) {
		return "container";
	}
	
	public boolean isSubscribed() throws InternalException, CloudException {
        BlobStoreContext ctx = provider.getCloudFilesContext();
        
        try {
            CloudFilesClient client = CloudFilesClient.class.cast(ctx.getProviderSpecificContext().getApi());
            
            client.listCDNContainers(ListCdnContainerOptions.NONE);
            return true;
        }
        catch( RuntimeException e ) {
            if( e.getMessage().contains("authorized") ) {
                return false;
            }
            logger.warn("Unable to determine CDN subscription status in Rackspace for " + provider.getContext().getAccountNumber() + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new CloudException(e);
        }
        finally {
            ctx.close();                        
        }
	}
	
	public Collection<Distribution> list() throws InternalException, CloudException {
	    logger.debug("enter - list()");
	    try {
	        ArrayList<Distribution> distributions = new ArrayList<Distribution>();
            Collection<ContainerCDNMetadata> list;
	        
	        try {
	            BlobStoreContext ctx = provider.getCloudFilesContext();
	            
	            try {
	                CloudFilesClient client = CloudFilesClient.class.cast(ctx.getProviderSpecificContext().getApi());
	                list = client.listCDNContainers(ListCdnContainerOptions.NONE);
	            }
	            catch( RuntimeException e ) {
	                e.printStackTrace();
	                throw new CloudException(e);
	            }
	            finally {
	                ctx.close();                        
	            }
	        }
	        catch( RuntimeException e ) {
	            logger.error("Could not list distributions in Rackspace: " + e.getMessage());
	            e.printStackTrace();
	            throw new CloudException(e);
	        }
	        for( ContainerCDNMetadata item : list ) {
	            if( item != null ) {
	                Distribution distribution = toDistribution(item);
	                
	                if( distribution != null ) {
	                    distributions.add(distribution);
	                }
	            }
	        }
	        return distributions;
	    }
	    finally {
	        logger.debug("exit - list()");
	    }
	}

	public void update(String distributionId, String name, boolean active, String ... cnames) throws InternalException, CloudException {
        logger.debug("enter - update(String, String, boolean, String ...)");
        try {
            Distribution distribution = getDistribution(distributionId);
            
            if( active ) {
                if( distribution == null ) {
                    create(name, name, active, cnames);
                }
                else if( !distribution.isDeployed() ) {
                    try {
                        BlobStoreContext ctx = provider.getCloudFilesContext();
                        
                        try {
                            CloudFilesClient client = CloudFilesClient.class.cast(ctx.getProviderSpecificContext().getApi());
                            client.enableCDN(distribution.getLocation());
                        }
                        catch( RuntimeException e ) {
                            logger.error("Could not enable distribution '" + distribution.getLocation() + "': " + e.getMessage());
                            e.printStackTrace();
                            throw new CloudException(e);
                        }
                        finally {
                            ctx.close();                        
                        }
                    }
                    catch( RuntimeException e ) {
                        logger.error("Could not enable " + distribution.getProviderDistributionId() + ": " + e.getMessage());
                        e.printStackTrace();
                        throw new CloudException(e);
                    }
                }
            }
            else if( distribution != null && distribution.isDeployed() || distribution.isActive() ) {
                try {
                    BlobStoreContext ctx = provider.getCloudFilesContext();
                    
                    try {
                        CloudFilesClient client = CloudFilesClient.class.cast(ctx.getProviderSpecificContext().getApi());
                        client.disableCDN(distribution.getLocation());
                    }
                    catch( RuntimeException e ) {
                        logger.error("Could not disable distribution '" + distribution.getLocation() + "': " + e.getMessage());
                        e.printStackTrace();
                        throw new CloudException(e);
                    }
                    finally {
                        ctx.close();                        
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Failed to disable " + distribution.getProviderDistributionId() + ": " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
            }
        }
        finally {
            logger.debug("exit - update(String, String, boolean, String ...)");            
        }
	}

	private Distribution toDistribution(ContainerCDNMetadata rackspaceDistribution) {
	    Distribution distribution = new Distribution();
	    
        distribution.setName(rackspaceDistribution.getName());
	    distribution.setDnsName(rackspaceDistribution.getCDNUri().getHost());
	    distribution.setActive(rackspaceDistribution.isCDNEnabled());
	    distribution.setDeployed(rackspaceDistribution.isCDNEnabled());
	    distribution.setLocation(rackspaceDistribution.getName());
	    distribution.setProviderDistributionId(rackspaceDistribution.getName());
	    return distribution;
	}
}
